<?php

# Утилитные функции
# (c) VitaliF 2006+

$AutoloadClasses['DatabaseMysql'] = dirname(__FILE__).'/DatabaseMysql.php';
$AutoloadClasses['LinguaStemRu'] = dirname(__FILE__).'/stem_ru.php';
$AutoloadClasses['VMXTemplate'] = dirname(__FILE__).'/template.php';
$AutoloadClasses['Mailer'] = dirname(__FILE__).'/mailing.php';

function __autoload($class_name)
{
    global $AutoloadClasses;
    if (isset($AutoloadClasses[$class_name]))
    {
        require_once $AutoloadClasses[$class_name];
        return true;
    }
    return false;
}

class PopulatedRow implements ArrayAccess
{
    public $row;
    public function __construct($row)
    {
        $this->row = $row;
    }
    public function offsetExists($k)
    {
        return array_key_exists($k, $this->row) || self::$calckeys[$k];
    }
    public function offsetUnset($k)
    {
        unset($this->$k);
    }
    public function offsetSet($k, $v)
    {
        if (array_key_exists($k, $this->row))
            $this->row[$k] = $v;
        else
            $this->$k = $v;
    }
    public function offsetGet($k)
    {
        if (array_key_exists($k, $this->row))
            return $this->row[$k];
        if (!is_null($this->$k))
            return $this->$k;
        $this->populate($k);
        return $this->$k;
    }
    public function populate($k)
    {
    }
}

function normalize_domain($st)
{
    $dom = '';
    if (!empty($st))
    {
        if (($p = strpos ($st, '://')) !== false)
            $dom = substr ($st, $p+3);
        $dom = strtolower ($dom);
        if (preg_match ('#^www[0-9]*\.#is', $dom, $m))
            $dom = substr ($dom, strlen ($m[0]));
        if (($p = strpos ($st, '/')) !== false)
            $dom = substr ($st, 0, $p);
    }
    return $dom;
}

function normalize_link($l, $pre)
{
    if (!empty ($l) && !empty ($pre))
    {
        if (str_starts_with($l, './'))
            $l = substr($l, 2);
        if (!strpos($l, '://'))
        {
            if ($l{0} != '/')
            {
                if ($pre{strlen($pre)-1} != '/')
                    $pre .= '/';
                $l = $pre . $l;
            }
            elseif (($sch = strpos($pre, '://')) !== false)
                $l = substr($pre, 0, strpos($pre, '/', $sch+3)) . $l;
        }
        return $l;
    }
    return $l;
}

function strlimit($str, $maxlen)
{
    if (!$maxlen || $maxlen < 1 || mb_strlen($str) <= $maxlen)
        return $str;
    $str = mb_substr($str, 0, $maxlen);
    $p = mb_strrpos($str, ' ');
    if (!$p || ($pt = mb_strrpos($str, "\t")) > $ps)
        $p = $pt;
    if ($p)
        $str = mb_substr($str, 0, $p);
    return $str . '...';
}

function html_pbr($str)
{
    return nl2br(htmlspecialchars($str));
}

function html_strip_pbr($str)
{
    global $allowed_html;
    if (!($a = $allowed_html))
        $a = '<div> <span> <a> <b> <i> <u> <p> <h1> <h2> <h3> <h4> <h5> <h6> <strike> <strong> <small> <big> <blink> <center> <ol> <pre> <sub> <sup> <font> <br> <table> <tr> <td> <th> <tbody> <tfoot> <thead> <tt> <ul> <li> <em> <img> <marquee>';
    return nl2br(strip_tags($str, $a));
}

function html_strip($str)
{
    global $allowed_html;
    if (!($a = $allowed_html))
        $a = '<div> <span> <a> <b> <i> <u> <p> <h1> <h2> <h3> <h4> <h5> <h6> <strike> <strong> <small> <big> <blink> <center> <ol> <pre> <sub> <sup> <font> <br> <table> <tr> <td> <th> <tbody> <tfoot> <thead> <tt> <ul> <li> <em> <img> <marquee>';
    return strip_tags($str, $a);
}

function unhtmlentity($s)
{
    $s = html_entity_decode($s);
    preg_match_all('/&#([0-9]*);/', $s, $ce, PREG_SET_ORDER | PREG_OFFSET_CAPTURE);
    $c = count($ce);
    for ($i = $c-1; $i >= 0; $i--)
        $s = substr($s, 0, $ce[$i][0][1]) . chr($ce[$i][1][0]) . substr($s, $ce[$i][0][1]+strlen($ce[$i][0][0]));
    return $s;
}

function process_array($arr, $kfunc, $vfunc)
{
    $na = array();
    foreach ($arr as $k => $v)
    {
        if (is_callable($kfunc))
            $k = $kfunc($k);
        if (is_callable($vfunc))
            $v = $vfunc($v);
        $na [$k] = $v;
    }
    return $na;
}

/**
 * Функция разворачивает массив во вложенный хеш
 * Например [[0,1],[0,2],[1,1]] -> [0:[1:[[0,1]],2:[[0,2]]],1:[1:[[1,1]]]]
 */
function array_hashtree($a, $ka = NULL, $value = NULL)
{
    $r = array();
    if (!is_array($ka))
        $ka = array($ka);
    foreach ($a as $ca)
    {
        $c = &$r;
        if (!is_null($ka))
        {
            foreach ($ka as $k)
            {
                $k = $ca[$k];
                if (!array_key_exists($k, $c))
                    $c[$k] = array();
                $c = &$c[$k];
            }
        }
        else
        {
            foreach ($ca as $k)
            {
                if (!array_key_exists($k, $c))
                    $c[$k] = array();
                $c = &$c[$k];
            }
        }
        if (is_null($value))
            $c[] = $ca;
        else
            $c = $value;
    }
    return $r;
}

/* Уложить ширину и высоту изображения пропорционально в рамки maxw/maxh */
function fit_width_height($cw, $ch, $maxw, $maxh)
{
    if ($maxw <= 0 && $maxh <= 0 || $cw <= 0 || $ch <= 0)
        return array($cw, $ch);
    if ($maxw > 0 && ($maxh <= 0 || $maxh/$ch >= $maxw/$cw))
    {
        $nw = $maxw;
        $nh = $ch * $maxw / $cw;
    }
    elseif ($maxh > 0 && ($maxw <= 0 || $maxh/$ch < $maxw/$cw))
    {
        $nw = $cw * $maxh / $ch;
        $nh = $maxh;
    }
    return array(intval($nw), intval($nh));
}

function convertBMP($from, $to, $quality = 90)
{
    static $class;
    $fp = fopen($from, "rb");
    if (!$fp)
    {
        return false;
    }
    $sign = fread($fp, 2);
    fclose($fp);
    if ($sign != 'BM')
    {
        return $from;
    }
    $class = class_exists('Gmagick') ? 'Gmagick' : 'Imagick';
    try
    {
        $im = new $class();
        $im->readImage($from);
        $im->setCompressionQuality($quality);
        $im->writeImage($to);
    }
    catch (Exception $e)
    {
        return $from;
    }
    return $to;
}

function loadimagebyext($from)
{
    if (strrpos($from, '.') !== false)
    {
        $ext = strtolower(substr($from, strrpos($from, '.')+1));
        if (function_exists("imagejpeg") && ($ext == 'jpg' || $ext == 'jpeg' || $ext == 'jfif' || $ext == 'jpe'))
            $img = @imagecreatefromjpeg($from);
        elseif (function_exists("imagepng") && ($ext == 'png'))
            $img = @imagecreatefrompng($from);
        elseif (function_exists("imagegif") && ($ext == 'gif'))
            $img = @imagecreatefromgif($from);
    }
    return $img;
}

function saveimagebyext($img, $to)
{
    global $default_jpeg_quality;
    $qual = $default_jpeg_quality;
    if (!$qual)
        $qual = 85;
    if (strrpos($to, '.') !== false)
    {
        $ext = strtolower(substr($to, strrpos($to, '.')+1));
        if ($ext == 'jpg' || $ext == 'jpeg' || $ext == 'jfif' || $ext == 'jpe')
            @imagejpeg($img, $to, $qual);
        elseif ($ext == 'png')
            @imagepng($img, $to);
        elseif ($ext == 'gif')
            @imagegif($img, $to);
        else
            return false;
        return $to;
    }
    return false;
}

/**
 * Масштабирование изображения, находящегося в файле $filename
 * Поддерживаются JPG, GIF, и PNG, в зависимости от возможностей PHP
 * В $maxw и $maxh возвратит результаты.
 */
function scaleimage ($from, $to, &$maxw, &$maxh, $ret_after_size = false)
{
    if (is_string($from))
        $img = loadimagebyext($from);
    else
        $img = $from;
    if (!$img)
        return false;
    $cw = imagesx($img);
    $ch = imagesy($img);
    $maxw = intval($maxw);
    $maxh = intval($maxh);
    if (($maxw <= 0 || $cw <= $maxw) && ($maxh <= 0 || $ch <= $maxh))
    {
        $maxw = $cw;
        $maxh = $ch;
        copy($from, $to);
        @imagedestroy($img);
        return true;
    }
    list($nw, $nh) = fit_width_height($cw, $ch, $maxw, $maxh);
    $thumb = imagecreatetruecolor($nw, $nh);
    $f = function_exists('imagecopyresampled') ? 'imagecopyresampled' : 'imagecopyresized';
    call_user_func($f, $thumb, $img, 0, 0, 0, 0, $nw, $nh, $cw, $ch);
    imagedestroy($img);
    $r = saveimagebyext($thumb, $to);
    imagedestroy($thumb);
    if ($ret_after_size)
    {
        $maxw = $nw;
        $maxh = $nh;
    }
    else
    {
        $maxw = $cw;
        $maxh = $ch;
    }
    return $r;
}

/* обрезание до нужного соотношения сторон $aspect = X/Y
   и потом - масштабирование до ширины $forcewidth, если она задана */
function centercropimage($from, $to, $aspect, $forcewidth = 0)
{
    $img = loadimagebyext($from);
    if (!$img)
        return false;
    $cw = imagesx($img);
    $ch = imagesy($img);
    $ca = $cw/$ch;
    if ($ca > $aspect)
    {
        $nw = $aspect*$ch;
        $nh = $ch;
    }
    else
    {
        $nw = $cw;
        $nh = $cw/$aspect;
    }
    $posx = intval(($cw-$nw)/2);
    $posy = intval(($ch-$nh)/2);
    $cw = $nw;
    $ch = $nh;
    if ($forcewidth > 0)
    {
        $nh = $nh*$forcewidth/$nw;
        $nw = $forcewidth;
    }
    $thumb = imagecreatetruecolor($nw, $nh);
    $f = function_exists('imagecopyresampled') ? 'imagecopyresampled' : 'imagecopyresized';
    call_user_func($f, $thumb, $img, 0, 0, $posx, $posy, $nw, $nh, $cw, $ch);
    imagedestroy($img);
    $r = saveimagebyext($thumb, $to);
    imagedestroy($thumb);
    return $r;
}

function get_image_size($from, &$w, &$h)
{
    if (strrpos ($from, '.') !== false)
    {
        $ext = strtolower (substr ($from, strrpos ($from, '.')+1));
        if (function_exists ("imagejpeg") && ($ext == 'jpg' || $ext == 'jpeg' || $ext == 'jfif' || $ext == 'jpe'))
            $img = @imagecreatefromjpeg ($from);
        elseif (function_exists ("imagepng") && ($ext == 'png'))
            $img = @imagecreatefrompng ($from);
        elseif (function_exists ("imagegif") && ($ext == 'gif'))
            $img = @imagecreatefromgif ($from);
        if ($img)
        {
            $w = imagesx ($img);
            $h = imagesy ($img);
            imagedestroy ($img);
            return true;
        }
    }
    return false;
}

function load_image ($filevarname, $newfilename, &$imgvar, &$thumbvar, &$imgw, &$imgh, $maxdim_config = '')
{
    global $local_path, $image_path, $file_creation_mode;
    $ret = false;

    $path = $image_path . $newfilename;
    if (($pos = strrpos ($path, '/')) !== false)
        $path = substr ($path, 0, $pos+1);
    else $path = '/';

    if (isset ($_FILES [$filevarname]['tmp_name']) &&
        $_FILES [$filevarname]['name'] != '')
    {
        $ext = strtolower (substr ($_FILES [$filevarname]['name'], strrpos ($_FILES [$filevarname]['name'], '.')+1));
        if ($ext == 'jpg' || $ext == 'jpeg' ||
            $ext == 'jpe' || $ext == 'jfif' ||
            $ext == 'png' || $ext == 'gif')
        {
            $fn_i = $image_path . $newfilename . '.' . $ext;
            $fn_t = $image_path . $newfilename . '_t.' . $ext;
            @unlink ($local_path . $fn_i); // !!! - предыдущее файло сотрётся
            @unlink ($local_path . $fn_t); // !!!
            if ((@move_uploaded_file ($_FILES [$filevarname]['tmp_name'], $local_path . $fn_i)))
            {
                if (!$imgw)
                    $imgw = 0+config_var ($maxdim_config.'img_maxw');
                if (!$imgh)
                    $imgh = 0+config_var ($maxdim_config.'img_maxh');
                if ($imgw <= 0)
                    $imgw = false;
                if ($imgh <= 0)
                    $imgh = false;
                if (is_numeric($file_creation_mode))
                    chmod($local_path . $fn_i, $file_creation_mode);
                if (scaleimage ($local_path . $fn_i, $local_path . $fn_t, $imgw, $imgh))
                {
                    $imgvar = $fn_i;
                    $thumbvar = $fn_t;
                    $ret = true;
                    if (is_numeric($file_creation_mode))
                        chmod($local_path . $fn_t, $file_creation_mode);
                }
            }
            else
                @unlink ($_FILES [$filevarname]['tmp_name']);
        }
        else
            @unlink ($_FILES [$filevarname]['tmp_name']);
    }
    return $ret;
}

function load_file ($filevarname, $newfilename)
{
    global $local_path, $file_creation_mode;
    $ret = false;

    $path = $newfilename;
    if (($pos = strrpos ($path, '/')) !== false)
        $path = substr ($path, 0, $pos+1);
    else $path = '/';

    if (isset ($_FILES [$filevarname]['tmp_name']))
    {
        $ext = strtolower (substr ($_FILES [$filevarname]['name'], strrpos ($_FILES [$filevarname]['name'], '.')+1));
        $fn = $newfilename . '.' . $ext;
        @unlink ($local_path . $fn);
        if ((@move_uploaded_file ($_FILES [$filevarname]['tmp_name'], $local_path . $fn)))
        {
            $ret = $fn;
            if (is_numeric($file_creation_mode))
                chmod($local_path . $fn, $file_creation_mode);
        }
        else
            @unlink ($_FILES [$filevarname]['tmp_name']);
    }
    return $ret;
}

function mkpath($path, $throw_error = true)
{
    if (is_dir($path) || @mkdir($path, 0777, true))
    {
        return true;
    }
    if ($throw_error)
    {
        $error = error_get_last();
        throw new Exception("Failed to create path $path: ".$error['message']);
    }
    return false;
}

function file_size($filename)
{
    global $local_path;
    $r = false;
    if ($filename{0} == '/' &&
        file_exists($local_path . $filename))
    {
        $r = filesize($local_path . $filename);
    }
    elseif (str_starts_with($filename, 'http://') ||
        str_starts_with($filename, 'ftp://'))
    {
        $a = @get_headers($filename, 1);
        if (isset($a['Content-Length']))
        {
            $r = $a['Content-Length'];
        }
    }
    return $r;
}

function file_size_string($bytes)
{
    $r = $bytes;
    if (is_numeric ($r))
    {
        if ($r >= 0 && $r < 1024)
            $r = $r . ' байт';
        elseif ($r >= 1024 && $r < 1024*1024)
            $r = sprintf ('%.2f Кб', $r/1024);
        elseif ($r >= 1024*1024 && $r < 1024*1024*1024)
            $r = sprintf ('%.2f Мб', $r/1024/1024);
        elseif ($r >= 1024*1024*1024)
            $r = sprintf ('%.2f Гб', $r/1024/1024/1024);
        elseif ($r < 0)
            $r = sprintf ('%.2f Гб', 2-($r/1024/1024/1024));
    }
    return $r;
}

/**
 * Получить данные от URL $url. Если $userpwd задана - тогда будет использована
 * HTTP аутентификация с логином и паролем, указанными в $userpwd как login:password.
 * Если данные являются gzip-сжатыми - они будут автоматически распакованы.
 */
function curl_get_contents($url, $options = array())
{
    global $local_path, $cache_path;

    $ch = curl_init();
    curl_setopt_array($ch, $options+array(
        CURLOPT_URL => $url,
        CURLOPT_ENCODING => 'gzip,deflate',
        CURLOPT_RETURNTRANSFER => 1,
    ));
    $data = @curl_exec($ch);
    curl_close($ch);

    return $data;
}

/* Получить некоторые строки по именам с автоматическим добавлением,
   если $add_unknown=true (например, теги) */
function get_tags($tags, $add_unknown = true, $table = 'tags', $id_field = 'id', $name_field = 'name')
{
    global $mysql, $_TagCache;
    if (!$tags)
        return $tags;
    $unk = array();
    foreach ($tags as $t)
    {
        $t = mb_strtolower($t, 'utf-8');
        if (!$_TagCache[$t])
            $unk[] = $t;
    }
    if ($unk)
    {
        $tagids = array_hashtree($mysql->select($table, "$id_field, LOWER($name_field) $name_field", array($name_field => $unk)), $name_field);
        $unk = array();
        foreach ($tags as $t)
        {
            if (!$tagids[mb_strtolower($t, 'utf-8')])
                $unk[] = array($name_field => $t);
        }
    }
    if ($unk && $add_unknown)
    {
        $mysql->update($table, $unk);
        $tagids = array_hashtree($mysql->select($table, "$id_field, LOWER($name_field) $name_field", array($name_field => $tags)), $name_field);
    }
    foreach ($tags as &$t)
    {
        $t = mb_strtolower($t, 'utf-8');
        if ($_TagCache[$t])
            $t = $_TagCache[$t];
        elseif ($tt = $tagids[$t])
        {
            $_TagCache[$t] = $tt[0];
            $t = $tt[0];
        }
    }
    return $tags;
}

function site_setcookie($name, $value, $expire, $secure = false, $httponly = false)
{
    global $Domain;
    list($dom, $path) = explode('/', $Domain, 2);
    $path = "/$path";
    return setcookie($name, $value, $expire ? time()+$expire : 0, $path, '.'.$dom, $secure, $httponly);
}

/**
 * Начало сессии :)
 * @param boolean $always Если true, то начинать сессию всегда, иначе - только если она существует
 * @param int $lifetime Необязательное время жизни сессии
 */
function session_begin($always = true, $lifetime = NULL)
{
    global $Domain, $LoginSession;
    if (isset($_COOKIE[$LoginSession]) || $always)
    {
        if ($lifetime === NULL)
        {
            $lifetime = 0;
        }
        session_name($LoginSession);
        list($dom, $path) = explode('/', $Domain, 2);
        $path = "/$path";
        $dom = (strtolower($dom) == 'localhost' ? '' : ".$dom");
        session_set_cookie_params($lifetime, $path, $dom);
        return @session_start();
    }
    return false;
}

function session_terminate()
{
    $_SESSION = array();
    $params = session_get_cookie_params();
    if (session_begin())
    {
        session_destroy();
    }
    setcookie(
        session_name(), '', -1, $params['path'], $params['domain'],
        $params['secure'], $params['httponly']
    );
}

/**
 * Определение нужного количества итераций блока $block на основе:
 * sql-запроса $sql ($row[0] это ID и $row[1] это NAME) и
 * массива $arr (массив сначала).
 * Функция предназначена для присваивания блоков <select>...</select>
 * и присвоит "выделено" той опции, которая задана ID-ом $sel_id
 */
function option_block($sql, $arr, $sel_id = NULL)
{
    global $mysql;
    $vars = array();
    if (is_array ($arr))
        foreach ($arr as $id => $name)
            $vars[] = array(
                'ID'   => $id,
                'NAME' => $name,
                'SEL'  => !is_null($sel_id) && $id == $sel_id ? 'selected' : '',
            );
    if (is_string($sql) && ($rows = $mysql->get_rows($sql)))
    {
        foreach ($rows as $row)
        {
            $vars[] = array(
                'ID'   => $row[0],
                'NAME' => $row[1],
                'SEL'  => !is_null($sel_id) && $row[0] == $sel_id ? 'selected' : '',
            );
        }
    }
    return $vars;
}

/**
 * Данные постраничной навигации внутри $allcount элементов по
 * $perpage на странице, при том условии что текущий первый отображённый - это $start.
 * Ссылка на страницу будет App::url($component, $args+array(page => НОМЕР_СТРАНИЦЫ)).
 * Если задано $eps, то отобразится не более 2*$eps соседних страниц.
 */
function page_nav3($total, $perpage, $curpage, $component, $args = array(), $eps = NULL, $metanext = 'METAS', $pagearg = 'page')
{
    if ($perpage <= 0 && $total > 0)
        $perpage = $total;
    if ($perpage <= 0)
        $perpage = 1;
    $pagecount = ceil($total / $perpage);
    if ($curpage >= $pagecount)
        $curpage = $pagecount-1;
    if ($curpage < 0)
        $curpage = 0;

    $vars = array (
        'total'      => $total,
        'perpage'    => $perpage,
        'list_start' => min($total, $curpage*$perpage+1),
        'page_count' => $pagecount,
        'list_end'   => min($total, $curpage*$perpage+$perpage),
        'page'       => array(),
        'current_page' => $curpage+1,
    );
    if ($pagecount < 2)
    {
        return $vars;
    }

    if ($curpage > 0)
    {
        $args[$pagearg] = $curpage-1;
        $vars['page'][] = array (
            'name' => '<',
            'href' => App::url($component, $args),
            'nc'   => true,
        );
    }
    if (is_null($eps))
        $eps = config_var('page_nav_eps', 10);
    if ($eps && $eps > 0)
    {
        $ppend = min($pagecount, $curpage + $eps + 1);
        for ($i = max(0, $curpage - $eps); $i < $ppend; $i++)
        {
            $args[$pagearg] = $i;
            $vars['page'][] = array(
                'name' => $i+1,
                'href' => App::url($component, $args),
                'nc'   => $i != $curpage,
            );
        }
    }
    else
    {
        for ($i = 0; $i < $pagecount; $i++)
        {
            $args[$pagearg] = $i;
            $vars['page'][] = array(
                'name' => $i+1,
                'href' => App::url($component, $args),
                'nc'   => $i != $curpage,
            );
        }
    }
    if ($curpage < $pagecount-1)
    {
        $args[$pagearg] = $curpage+1;
        if ($metanext)
        {
            global $template;
            $template->tpldata[$metanext] .= '<link rel="next" title="Следующая страница" href="'.App::url($component, $args).'"/>';
        }
        $vars['page'][] = array(
            'name' => '>',
            'href' => App::url($component, $args),
            'nc'   => true,
        );
    }
    return $vars;
}

if (!function_exists("quoted_printable_encode"))
{
    define('QP_LINE_LENGTH', 75);
    define('QP_LINE_SEPARATOR', "\r\n");

    function quoted_printable_encode($string, $encodedWord = false)
    {
        if(!preg_match('//u', $string)) {
            throw new Exception('Input string is not valid UTF-8');
        }

        static $wordStart = '=?UTF-8?Q?';
        static $wordEnd = '?=';
        static $endl = QP_LINE_SEPARATOR;

        $lineLength = $encodedWord
            ? QP_LINE_LENGTH - strlen($wordStart) - strlen($wordEnd)
            : QP_LINE_LENGTH;

        $string = $encodedWord
            ? preg_replace('~[\r\n]+~', ' ', $string)    // we need encoded word to be single line
            : preg_replace('~\r\n?~', "\n", $string);    // normalize line endings
        $string = preg_replace('~[\x00-\x08\x0B-\x1F]+~', '', $string);    // remove control characters

        $output = $encodedWord ? $wordStart : '';
        $charsLeft = $lineLength;

        $chr = isset($string{0}) ? $string{0} : null;
        $ord = ord($chr);

        for ($i = 0; isset($chr); $i++) {
            $nextChr = isset($string{$i + 1}) ? $string{$i + 1} : null;
            $nextOrd = ord($nextChr);

            if (
                $ord > 127 or    // high byte value
                $ord === 95 or    // underscore "_"
                $ord === 63 && $encodedWord or    // "?" in encoded word
                $ord === 61 or    // equal sign "="
                // space or tab in encoded word or at line end
                $ord === 32 || $ord === 9 and $encodedWord || !isset($nextOrd) || $nextOrd === 10
            ) {
                $chr = sprintf('=%02X', $ord);
            }

            if ($ord === 10) {    // line feed
                $output .= $endl;
                $charsLeft = $lineLength;
            } elseif (
                strlen($chr) < $charsLeft or
                strlen($chr) === $charsLeft and $nextOrd === 10 || $encodedWord
            ) {    // add character
                $output .= $chr;
                $charsLeft-=strlen($chr);
            } elseif (isset($nextOrd)) {    // another line needed
                $output .= $encodedWord
                    ? $wordEnd . $endl . "\t" . $wordStart . $chr
                    : '=' . $endl . $chr;
                $charsLeft = $lineLength - strlen($chr);
            }

            $chr = $nextChr;
            $ord = $nextOrd;
        }

        return $output . ($encodedWord ? $wordEnd : '');
    }
}

function litsplit($re, $s)
{
    $r = array();
    $s = trim($s);
    $lit = "\'(?:[^\'\\\\]+|\\.|\'\')+\'|\"(?:[^\"\\\\]+|\\.|\"\")+\"";
    $offset = 0;
    $re = "/((?:$lit|.+?)*?)($re)/is";
    while (preg_match($re, $s, $m, PREG_OFFSET_CAPTURE, $offset))
    {
        $r[] = $m[1][0];
        $offset = $m[0][1] + strlen($m[0][0]);
    }
    $s = substr($s, $offset);
    if (strlen($s))
        $r[] = $s;
    return $r;
}

function probe_video($fn)
{
    global $VideoProbe;
    if (!$VideoProbe || !$VideoProbe['cmd'] || !$VideoProbe['size'] || !$VideoProbe['duration'])
        die(__FUNCTION__.": \$VideoProbe=array('cmd'=>'shell command','size'=>'size regexp','duration'=>'duration regexp') isn't configured yet");
    $cmd = $VideoProbe['cmd'];
    $cmd = str_replace('$input', $fn, $cmd);
    $out = shell_exec($cmd);
    $db = array();
    if (preg_match($VideoProbe['size'], $out, $m))
    {
        $db['width'] = $m[1];
        $db['height'] = $m[2];
    }
    if (preg_match($VideoProbe['duration'], $out, $m))
    {
        $dur = $m[1];
        $colon = explode(':', $dur);
        if (count($colon) > 1)
        {
            $dur = 0;
            for ($mul = 1, $i = 0; $i < count($colon); $i++, $mul *= 60)
                $dur += $colon[count($colon)-1-$i] * $mul;
        }
        $db['duration'] = $dur;
    }
    return $db;
}

function extract_video_frame($from, $time, $w, $h, $to)
{
    global $VideoConverter;
    if (!$VideoConverter)
        die(__FUNCTION.": \$VideoConverter = 'shell command' isn't configured yet");
    $cmd = $VideoConverter;
    $repl = array(
        '$input' => $from,
        '$position' => $time,
        '$output' => $to,
    );
    $cmd = str_replace(array_keys($repl), array_values($repl), $cmd);
    exec($cmd, $output, $exitcode);
    if ($exitcode != 0)
        return false;
    scaleimage($to, $to, $w, $h);
    return true;
}

function array_id_name($assoc)
{
    $r = array();
    foreach($assoc as $k => $v)
        $r[] = array('id' => $k, 'name' => $v);
    return $r;
}

function urandom($nbytes = 16, $hex = false)
{
    $pr_bits = '';
    // Unix/Linux platform?
    $fp = @fopen('/dev/urandom', 'rb');
    if ($fp !== FALSE)
    {
        $pr_bits = @fread($fp, $nbytes);
        @fclose($fp);
    }
    // MS-Windows platform?
    elseif (@class_exists('COM'))
    {
        // http://msdn.microsoft.com/en-us/library/aa388176(VS.85).aspx
        try
        {
            $com = new COM('CAPICOM.Utilities.1');
            if (method_exists($com, 'GetRandom'))
                $pr_bits = base64_decode($com->GetRandom($nbytes,0));
            else
            {
                $com = new COM('System.Security.Cryptography.RNGCryptoServiceProvider');
                if (method_exists($com, 'GetBytes'))
                    $pr_bits = base64_decode($com->GetBytes($nbytes));
            }
        }
        catch (Exception $ex)
        {
        }
    }
    if (!strlen($pr_bits))
        for ($i = 0; $i < $nbytes; $i++)
            $pr_bits .= chr(mt_rand(0, 255));
    if ($hex)
    {
        $pr_bits = unpack('H*', $pr_bits);
        $pr_bits = $pr_bits[1];
    }
    return $pr_bits;
}

// Convert all UTF8 characters except ASCII and &"'<>
// to HTML entities (&amp; &quot; &apos; &lt; &gt; &#xHEX;)
function htmlentities_utf8($s)
{
    $s = unpack('C*', mb_convert_encoding($s, 'UTF-16BE', 'UTF-8'));
    $r = '';
    for ($i = 1, $n = count($s); $i <= $n; $i += 2)
    {
        if (!$s[$i])
        {
            if ($s[$i+1] == 0x26)
                $r .= '&amp;';
            elseif ($s[$i+1] == 0x22)
                $r .= '&quot;';
            elseif ($s[$i+1] == 0x27)
                $r .= '&apos;';
            elseif ($s[$i+1] == 0x3C)
                $r .= '&lt;';
            elseif ($s[$i+1] == 0x3E)
                $r .= '&gt;';
            else
                $r .= chr($s[$i+1]);
        }
        else
            $r .= sprintf("&#x%x;", ($s[$i]<<8) + ($s[$i+1]));
    }
    return $r;
}

function load_config()
{
    global $mysql, $engine_configuration, $mysqlNoSiteConfig, $mysqlSiteConfigQuery;
    if ($mysqlNoSiteConfig)
        return;
    $engine_configuration = array();
    if ($mysqlSiteConfigQuery)
        $rows = $mysql->select($mysqlSiteConfigQuery, MS_LIST);
    else
        $rows = $mysql->select('config', array('name', 'value'), 1, array(), MS_LIST);
    foreach ($rows as $row)
        $engine_configuration[$row[0]] = $row[1];
}

function is_intval($v)
{
    return is_int($v) || is_string($v) && ctype_digit($v);
}

function config_var($name, $default = NULL)
{
    global $engine_configuration;
    if (!$engine_configuration)
        load_config();
    $v = @$engine_configuration[$name];
    if (!is_null($default) && !strlen($v) || is_intval($default) && $v <= 0)
        $v = $default;
    elseif (is_intval($default))
        $v = intval($v);
    return $v;
}
